#version 460 core

#define Scale vec3(.8, .8, .8)
#define K 19.19

// Heavily inspired by https://github.com/ImanolFotia/Epsilon-Engine/blob/master/bin/Release/shaders/SSR.frag
// Not a carbon copy though 

noperspective in vec2 UV;

//out vec3 color;

uniform sampler2D renderedSceneTex;		//Rendered Scene
uniform sampler2D positionTex;			//World space position
uniform sampler2D normalTex;			//World space normals
uniform sampler2D extraComponentsTex;	//x = metallic, y = roughness

uniform mat4 viewMat;
uniform mat4 projectionMat;
uniform mat4 viewProjMat;
uniform vec3 viewPos;

const float step = 0.15;								//initial step size
const float stepSizeGrowFactor = 1.02;					//increase step size by 3 percent each step. 
const float maxSteps = 90;								//maximum search steps. (seldomly reached)
const int numBinarySearchSteps = 10;					
const float reflectionSpecularFalloffExponent = 3.0;

int steps = 0;
int binSteps = 0;

vec3 PositionFromDepth(float depth);
vec3 BinarySearch(vec3 dir, vec3 hitCoord);
vec3 RayMarch(vec3 dir, inout vec3 hitCoord);
vec3 fresnelSchlick(float cosTheta, vec3 F0);
vec3 hash(vec3 a);

out vec3 outColor;				//maybe write to different output layout (default = 0)


vec3 viewPosition;
vec3 worldPosition;

vec3 fromLinear(vec3 linearRGB)
{
    bvec3 cutoff = lessThan(linearRGB, vec3(0.0031308));
    vec3 higher = vec3(1.055)*pow(linearRGB, vec3(1.0/2.4)) - vec3(0.055);
    vec3 lower = linearRGB * vec3(12.92);

    return mix(higher, lower, cutoff);
}

vec3 toLinear(vec3 sRGB)
{
    bvec3 cutoff = lessThan(sRGB, vec3(0.04045));
    vec3 higher = pow((sRGB + vec3(0.055))/vec3(1.055), vec3(2.4));
    vec3 lower = sRGB/vec3(12.92);

    return mix(higher, lower, cutoff);
}

void debugPrint(vec3 coords){
if (UV.x > 0.5){
		float c = steps / maxSteps;
		if (UV.x > 0.75){
			c = binSteps / numBinarySearchSteps;
		}
		outColor = vec3(c,1-c,0);
		if (coords.z < 0.1){
		outColor = vec3(0);
		}
	} 
}

void main(){

	vec2 MetalRough = texture2D(extraComponentsTex, UV).xy;
	float metallic = MetalRough.x;
	float roughness = MetalRough.y;

	//metallic = 0.8;
	//roughness = 0;

	vec4 normal = texture2D(normalTex, UV);
	vec3 worldNormal = normalize(normal.xyz);

	vec4 position = texture2D(positionTex, UV);
	worldPosition = position.xyz;

	vec3 viewVec = (worldPosition - viewPos);

	vec3 albedo = texture(renderedSceneTex, UV).rgb;
			
	//Fresnel
		vec3 F0		 = vec3(0.04);
		F0			 = mix(F0, albedo, metallic);
		vec3 Fresnel = fresnelSchlick(max(dot(worldNormal, viewVec), 0.0), F0);																	//Fresnel effect = reflectiveness increases with shallow angle to reflecting surface
	
	vec3 reflectionRay = normalize(reflect(viewVec, worldNormal));	
	vec3 hitPos = worldPosition;

	vec3 randomJitter = mix(vec3(0), vec3(hash(worldPosition)), roughness);																		//the rougher the surface the stronger the jitter
	vec3 coords = RayMarch((randomJitter + reflectionRay), hitPos);
	float wasHit = coords.z;

	vec2 dCoords = smoothstep(0.6, 1.0, abs(vec2(0.5,0.5) - coords.xy) * 2);																					//faze out edge artifacts, test this
	float screenEdgeFadeFactor = clamp(1 - (dCoords.x + dCoords.y), 0, 1);

	float reflectionFactor = pow(metallic, reflectionSpecularFalloffExponent)  * screenEdgeFadeFactor * /*max(0,dot(viewVec, reflectionRay)) **/ wasHit;		
	
	vec3 SSR = textureLod ( renderedSceneTex, coords.xy, 0).rgb * clamp(reflectionFactor, 0.0, 1.0) * Fresnel;									//maybe use LOD for better performance if needed
	
	outColor = albedo + SSR * metallic; //vec4(SSR, metallic
	outColor = toLinear(outColor);

	//debugPrint(coords);	
	//outColor = vec3(metallic);
	//outColor = albedo;
}

vec3 fresnelSchlick(float cosTheta, vec3 F0)
{
    return F0 + (1.0 - F0) * pow(1.0 - cosTheta, 5.0);
}

vec3 hash(vec3 a)
{
    a = fract(a * Scale);
    a += dot(a, a.yxz + K);
    return fract((a.xxy + a.yxx)*a.zyx);
}

float calcDeltaRayDepthTexDepth(vec2 projectedCoord, vec3 hitCoord){
		float distance_cam_texDepth	= length(viewPos - (textureLod(positionTex, projectedCoord.xy, 0)).xyz);		
		float distance_cam_rayStepDepth = length(viewPos - (hitCoord));		
		float difference_ray_depth_and_tex_depth = distance_cam_texDepth - distance_cam_rayStepDepth;
		return difference_ray_depth_and_tex_depth;
}

vec4 worldToCamSpace(vec4 projectedCoord){
		projectedCoord = viewProjMat * projectedCoord;
		projectedCoord.xy /= projectedCoord.w;
		projectedCoord.xy = projectedCoord.xy * 0.5 + 0.5;
		return projectedCoord;
}


vec3 RayMarch(vec3 dir, inout vec3 hitCoord)
{
	dir *= step; 
	vec4 projectedCoord;
	float difference_ray_depth_and_tex_depth;	
	for (int i = 0; i < maxSteps; i++){
		dir *= 1.03;
		hitCoord += dir;

		projectedCoord = worldToCamSpace(vec4(hitCoord, 1));

		if (projectedCoord.x > 1 || projectedCoord.x < 0 || projectedCoord.y > 1 || projectedCoord.y < 0){
			return vec3(0);
		}

		difference_ray_depth_and_tex_depth = calcDeltaRayDepthTexDepth(projectedCoord.xy, hitCoord);

		if ((difference_ray_depth_and_tex_depth) <= 0.0)																//below 0 = hit, far below 0, dont hit (would cause infinite long reflections)
		{			
			vec3 binSearchRes = BinarySearch(dir, hitCoord);
			if (binSearchRes.z > 0.5){
			//outColor = vec3(0,1,0);
				return binSearchRes;
			}			
		}
		steps++;																				//so we can do some checking how many steps are needed for which surface, we can use colors yay
	}
	//outColor = vec3(0,1,0);

	return vec3(0,0,0);
}

vec3 BinarySearch(vec3 dir, vec3 hitCoord)
{
	float difference_ray_depth_and_tex_depth;
    vec4 projectedCoord;
 
    for(int i = 0; i < numBinarySearchSteps; i++)
    {
		binSteps++;
		projectedCoord = worldToCamSpace(vec4(hitCoord, 1));
		difference_ray_depth_and_tex_depth = calcDeltaRayDepthTexDepth(projectedCoord.xy, hitCoord);

        dir *= 0.5;
        if(difference_ray_depth_and_tex_depth > 0.0)
            hitCoord += dir;
        else
            hitCoord -= dir;
    }

	//ray may now actually not hit anymore, if the last step moved us out of the hit geometry, go another step back into it.
	projectedCoord = worldToCamSpace(vec4(hitCoord, 1));
	difference_ray_depth_and_tex_depth = calcDeltaRayDepthTexDepth(projectedCoord.xy, hitCoord);
	if(difference_ray_depth_and_tex_depth > 0.0)
            hitCoord += dir;

	projectedCoord = worldToCamSpace(vec4(hitCoord, 1));

	if (calcDeltaRayDepthTexDepth(projectedCoord.xy, hitCoord) > -0.5)
		{
			return vec3(projectedCoord.xy,1);
		} else {
			return vec3(0,0,0);
		}
}

